"""
Modul mit verschiedensten allgemeinen Hilfsklassen
"""

# coding=utf-8
import openpyxl
from collections import Counter, OrderedDict
from tools.text import PrettyText
from tools.coding import Result
from tools.printing import SmartPrinter

class MultiSheet:
    """
    MultiSheet Datenmodell
    """

    # --- Konstruktor ---

    def __init__(self, smartprinter = None, **kwargs):
        self.__smartprinter = smartprinter if smartprinter is not None else SmartPrinter()

        self.__sWbName = kwargs["name"] if "name" in kwargs else 'Multisheet'
        self.__odicWb = OrderedDict()

        self.__dicIndex_names = {}
        self.__dicIndex = {}
        self.__sOpenSheet = None

        self.__lstIterSheetContent = None

        self.__bFinalized = False
        self.__result = Result("Multi Sheet")

    def __str__(self):
        return self.getInfo()

    # --- Iterator ---

    def __iter__(self):
        return self

    def __call__(self):
        try:
            self.__lstIterSheetContent = iter(self.__odicWb[self.getOpenSheetName()]["content"])
        except:
            pass
        return self

    def __next__(self):
        if self.__lstIterSheetContent is None:
            raise StopIteration
        lstRow = next(self.__lstIterSheetContent)
        #dicRow = dict(zip(self.getHeader(self.__sIterSheetName), lstRow))
        return lstRow

    # --- Info ---

    def getInfo(self, **kwargs):
        """
        Info

        kwargs:
            full (bool): gib den Tabelleninhalt mit aus
        return:
            info (str)
        """
        bFull = kwargs["full"] if "full" in kwargs else False
        sOut = ""
        sOut += self.getInfo_stat()
        if bFull:
            sOut += "\n"
            sOut += self.getInfo_table_ascii()
        return sOut

    def getInfo_stat(self):
        sOut = ""
        sOut += f"* Workbook\n"
        sOut += f"Name: {self.getWorkbookName()}\n"
        sOut += f"Rows: {self.getNoOfRows()}\n"
        sOut += f"Open sheet: {self.getOpenSheetName()}\n"
        if self.isFinalized():
            sOut += "Finalized: yes\n"
            if self.hasIndex():
                sIndex_info = ""
                if self.isIndexEmpty():
                    sIndex_info = "empty map"
                else:
                    sIndex_info = "working"
                sOut += f"Index ({sIndex_info}):\n"
                for sSheet in self.getSheetNames():
                    if self.hasIndex(sSheet):
                        sOut += f"{sSheet} [{PrettyText.lst2strC(self.getIndex(sSheet))}]\n"
            else:
                sOut += "Index: not set\n"
        else:
            sOut += "Finalized: no\n"
        sOut += "Overview:\n"
        lstTable = [["sheet", "rows"]]
        for sSheet, iLen in self.getStat().items():
            lstTable.append([sSheet, iLen])
        sOut += PrettyText.table(lstTable) + "\n"
        if self.__result.isAnyMessage():
            sOut += "\n"
            sOut += self.__result.getErrWarnInfoStringHeader()
            sOut += self.__result.getErrWarnInfoString(short=True)
            sOut += self.__result.getErrWarnInfoStringFooter()
        return sOut

    def getInfo_table_ascii(self):
        sOut = ""
        for sSheet, dicData in self.__odicWb.items():
            sOut += sSheet + " (#" + str(self.getNoOfRows(sSheet)) + ")\n"
            lstHeader = dicData["header"]
            lstContent = dicData["content"]
            lstTable = [lstHeader]
            lstTable += lstContent
            sOut += PrettyText.table(lstTable, True) + "\n\n"
        return sOut

    # --- Finalize ---

    def finalize(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        iNow = 0
        iTotal = self.getNoOfRows()
        for sSheet in self.getSheetNames():
            iRows = self.getNoOfRows(sSheet)
            lstHeader = self.getColNames(sSheet)
            iHeader_len = len(lstHeader)
            if lstHeader is None: lstHeader = []
            for iRow in range(iRows):
                progress(iNow, iTotal, 100, proz=True, cli=True)
                iNow += 1
                lstRow = self.getRow(sSheet, row=iRow, mode="list_ref") # Hole Zeile als Referenz
                iRow_len = len(lstRow)
                # Header/Content prüfen
                if iRow_len > iHeader_len:
                    self.__result.addErr("Row longer than header", row=iRow, sheet=sSheet,sys="MultiSheet Data", kat="Finalize")
                elif iRow_len < iHeader_len:
                    for i in range(iHeader_len - iRow_len):
                        lstRow.append(None) # Auffüllen (INPLACE)
                    self.__result.addInfo("Row shorter than header -> filled with None", row=iRow, sheet=sSheet,sys="MultiSheet Data", kat="Finalize")
                # Index erstellen und prüfen
                if self.hasIndex(sSheet):
                    lstIdx = self.getIndex(sSheet)
                    sIndex = ""
                    lstValues = []
                    val = None
                    for sHeader_index in lstIdx:
                        resVal = self.getValue(sSheet, sHeader_index, row=iRow)
                        if resVal.isErr():
                            self.__result.addErr("No index data found for header", add=sHeader_index, sheet=sSheet,sys="MultiSheet Data", kat="Indexing")
                        else:
                            val = resVal.getData("value")
                            lstValues.append(val)
                    sIndex = PrettyText.lst2str(lstValues, "|")
                    if sIndex in self.__dicIndex[sSheet]:
                        self.__result.addErr("Index already existing", add=sIndex, sheet=sSheet, sys="MultiSheet Data",kat="Indexing")
                    else:
                        self.__dicIndex[sSheet][sIndex] = iRow
        self.__bFinalized = True

    # --- Create ---

    def addSheet(self, sSheet, lstHeaders, lstContent=None):
        res = Result("Add sheet to multisheet")
        if lstContent is None:
            lstContent = []
        if lstHeaders is None:
            lstHeaders = []
        if sSheet in [None, ""]:
            res.addErr("No sheet given", sys="MultiSheet Data", kat="Add sheet")
            return res
        if sSheet in self.__odicWb:
            res.addErr("Sheet already existing", sheet=sSheet, sys="MultiSheet Data", kat="Add sheet")
            return res
        if lstHeaders is not None and len(lstHeaders) == 0:
            res.addErr("No headers given for for Sheet", sheet=sSheet, sys="MultiSheet Data", kat="Add sheet")
            return res
        if type(lstContent) is not list:
            res.addErr("Content is no list", sheet=sSheet, sys="MultiSheet Data", kat="Add sheet")
            return res
        self.__odicWb[sSheet] = {"header": lstHeaders, "content": lstContent}
        return res

    def addRow(self, sSheet, dat_in):
        res = Result("Add row to multisheet")
        if type(dat_in) is list:
            return self.__addRow_list(sSheet, dat_in)
        elif type(dat_in) is dict:
            return self.__addRow_dic(sSheet, dat_in)
        else:
            res.addErr("Row wasn't either a list nor dictionay", sheet=sSheet, sys="MultiSheet Data", kat="Add row")
            return res
    def __addRow_list(self, sSheet, lstRow):
        res = Result("Add row to multisheet")
        if sSheet not in self.__odicWb:
            res.addErr("Sheet not found in Multisheet", add=sSheet, sys="MultiSheet Data", kat="Add row")
            return res
        try:
            iHeader = len(self.__odicWb[sSheet]["header"])
            iRow = len(lstRow)
            if iRow > iHeader:
                # Zeile ist länger als erlaubt (Header) -> Fehler
                res.addErr("Could not add row to sheet: row is to long", sheet=sSheet, sys="Multisheet", kat="Add row")
                return res
            elif iRow < iHeader:
                # Zeile ist kürzer als erlaubt (Header) -> Rest auffüllen mit None
                for i in range(iHeader-iRow):
                    lstRow.append(None) # Auffüllen
            self.__odicWb[sSheet]["content"].append(lstRow)
        except:
            res.addErr("Could not add row to sheet", sheet=sSheet, sys="Multisheet", kat="Add index")
        return res
    def __addRow_dic(self, sSheet, dicRow):
        res = Result("Add row to multisheet")
        if sSheet not in self.__odicWb:
            res.addErr("Sheet not found in Multisheet", add=sSheet, sys="MultiSheet Data", kat="Add row")
            return res
        lstRow = []
        for sCol in self.getColNames(sSheet):
            if sCol in dicRow:
                lstRow.append(dicRow[sCol])
                dicRow.pop(sCol)
            else:
                lstRow.append(None)
        if len(dicRow) > 0:
            res.addErr("Some headers not found in multisheet", add=PrettyText.dic2str(dicRow), sheet=sSheet, sys="MultiSheet Data", kat="Add row")
            return res
        else:
            return self.__addRow_list(sSheet, lstRow)

    # --- Setter ---

    def setOpenSheetName(self, sSheet):
        if sSheet in self.__odicWb:
            self.__sOpenSheet = sSheet
            return True
        else:
            return False

    def setOpenSheetIdx(self, iIdx):
        try:
            self.__sOpenSheet = list(self.__odicWb.keys())[iIdx]
            return True
        except:
            return False

    def setIndex(self, sSheet, lstIndex):
        if sSheet not in self.getSheetNames():
            self.__result.addErr("Sheet not found", add=sSheet, sys="Multisheet", kat="Set index")
            return False
        for sIndex in lstIndex:
            if sIndex not in self.getColNames(sSheet):
                self.__result.addErr("Column not found", add=sIndex, sys="Multisheet", kat="Set index")
                return False
        self.__dicIndex_names[sSheet] = lstIndex
        self.__dicIndex[sSheet] = {}
        return True

    # --- Getter ---

    def saveToExcel(self, sFile):
        res = Result()
        if sFile is None:
            res.addErr("No file given", sys="Multisheet", kat="Save to Excel")
            return res
        try:
            wb = openpyxl.Workbook()
            bFirstSheet = True
            for sSheet in self.getSheetNames():
                # Sheet
                if bFirstSheet:
                    ws = wb.active
                    ws.title = sSheet
                    bFirstSheet = False
                else:
                    ws = wb.create_sheet(title=sSheet)
                lstHeader, lstContent = self.getSheet(sSheet)
                # Header
                lstHeader_styled = []
                for head in lstHeader:
                    cell = openpyxl.cell.WriteOnlyCell(ws, value=head)
                    cell.font = openpyxl.styles.Font(bold=True)
                    lstHeader_styled.append(cell)
                ws.append(lstHeader_styled)
                # Daten
                for row in lstContent:
                    ws.append(row)
            wb.save(filename=sFile)
        except Exception as ex:
            res.addErr("Unexcepted problem occurred", add=str(ex), sys="Multisheet", kat="Save to Excel")
        return res

    # ---

    def getIndex(self, sSheet):
        return self.__dicIndex_names.get(sSheet)

    def getResult(self):
        return self.__result

    def isOk(self):
        return not self.isEmpty() and not self.isErr()

    def isErr(self):
        return self.__result.isErr()

    def hasIndex(self, sSheet=None):
        if sSheet is None:
            for sSheet, lstIndex in self.__dicIndex_names.items():
                if len(lstIndex) > 0: return True
            return False
        else:
            return self.getIndex(sSheet)

    def isEmpty(self):
        return len(self.__odicWb) == 0

    def getWorkbookName(self):
        return self.__sWbName

    def getOpenSheetName(self):
        if self.__sOpenSheet is not None:
            return self.__sOpenSheet
        try:
            return self.getSheetNames()[0]
        except:
            return None

    def getOpenSheetIdx(self):
        if self.__sOpenSheet is not None:
            return list(self.__odicWb.keys()).index(self.__sOpenSheet)
        else:
            return 0

    def getSheet(self, sSheet=None):
        if sSheet is None:
            sSheet = self.getOpenSheetName()
        if sSheet is None:
            return None, None
        return self.__odicWb[sSheet]["header"], self.__odicWb[sSheet]["content"]

    def getStat(self):
        dicStat = {}
        for sSheet, odicData in self.__odicWb.items():
            dicStat[sSheet] = len(odicData["content"])
        return dicStat

    def getNoOfRows(self, sSheetIn=None):
        iRows = 0
        for sSheet, odicData in self.__odicWb.items():
                if (sSheet == sSheetIn) or (sSheetIn is None):
                    iRows += len(odicData["content"])
        return iRows

    def getRow(self, sSheet, **kwargs):
        """
        Hole Zeile als Liste
        Kwargs:
            mode (str: dic, list, list_ref): Hole Zeile als dic, Liste (copy) oder Liste (Referenz)
            row (int):  Zeile
            index (list): Index
        Return:
            (list/dic/none)
        """
        sMode = kwargs["mode"] if "mode" in kwargs else "list"
        iRow = kwargs["row"] if "row" in kwargs else None
        lstIndex = kwargs["index"] if "index" in kwargs else None
        if lstIndex is not None and type(lstIndex) is list and iRow is None:
            sIndex = PrettyText.lst2str(lstIndex, "|")
            iRow = self.__dicIndex[sSheet].get(sIndex)
        if iRow is None or type(iRow) is not int: return None
        if sSheet not in self.getSheetNames(): return None
        if iRow < 0 or iRow > self.getNoOfRows(sSheet): return None
        if sMode not in ["dic", "list", "list_ref"]: return None
        if sMode == "dic":
            return self.__getRow_dic(sSheet, iRow)
        elif sMode == "list":
            return self.__getRow_list(sSheet, iRow, True)
        elif sMode == "list_ref":
            return self.__getRow_list(sSheet, iRow, False)
        else:
            return None
    def __getRow_list(self, sSheet, iRow, bCopy=True):
        if sSheet not in self.getSheetNames():
            return None
        if iRow < 0 or iRow > self.getNoOfRows(sSheet):
            return None
        else:
            try:
                lstRow = self.__odicWb[sSheet]["content"][iRow]
                if bCopy:
                    lstRow = list(lstRow)
                return lstRow
            except:
                return None
    def __getRow_dic(self, sSheet, iRow):
        if sSheet not in self.getSheetNames():
            return None
        if iRow < 0 or iRow > self.getNoOfRows(sSheet):
            return None
        else:
            lstValues = self.__odicWb[sSheet]["content"][iRow]
            lstHeader = self.getColNames(sSheet)
            dicRow = dict(zip(lstHeader, lstValues))
            return dicRow

    def getValue(self, sSheet, sHead, **kwargs):
        """
        Hole Wert
        Args:
            sSheet (str)
            sHead (str)
        Kwargs:
            row (int):  Zeile
            index (list): Index
        Returns:
            Value
        """
        res = Result("Get value from multisheet")
        iRow = kwargs["row"] if "row" in kwargs else None
        lstIndex = kwargs["index"] if "index" in kwargs else None
        dicErg = self.getRow(sSheet, mode="dic", row=iRow, index=lstIndex)
        sAdd = ""
        if iRow is not None:
            sAdd += f"row:{iRow}"
        if lstIndex is not None:
            if sAdd != "":
                sAdd += " "
            sAdd += f'index:{PrettyText.lst2str(lstIndex,"|")}'
        if dicErg is None:
            res.addErr("Value not found", add=sAdd, col=sHead, sheet=sSheet, sys="Multisheet", kat="Get value")
        elif sHead not in dicErg:
            res.addErr("Value not found", add=sAdd, col=sHead, sheet=sSheet, sys="Multisheet", kat="Get value")
        else:
            value = dicErg.get(sHead)
            res.addData("value", value)
        return res

    def setValue(self, sSheet, sHead, value, **kwargs):
        """
        Setze Wert
        Args:
            sSheet (str)
            sHead (str)
            value
        Kwargs:
            row (int):  Zeile
            index (list): Index
        Returns:
            (bool)
        """
        pn = self.__smartprinter.normal
        iRow = kwargs["row"] if "row" in kwargs else None
        lstIndex = kwargs["index"] if "index" in kwargs else None
        if lstIndex is not None and type(lstIndex) is list and iRow is None:
            sIndex = PrettyText.lst2str(lstIndex,"|")
            iRow = self.__dicIndex[sSheet].get(sIndex)
        if iRow is None or type(iRow) is not int: return None
        if sHead not in self.getColNames(sSheet): return None
        if sSheet not in self.getSheetNames(): return None
        if iRow < 0 or iRow > self.getNoOfRows(sSheet): return None
        try:
            iIdxHead = self.__odicWb[sSheet]["header"].index(sHead)
            row = self.__odicWb[sSheet]["content"][iRow] # Kann eine Liste oder ein Tupel (ro) sein
            # Konvertiere Zeile in Liste (rw)
            if type(row) is tuple:
                row = list(row)
            # Verändere Eintrag in row
            row[iIdxHead] = value
            # Tausche row in content durch
            self.__odicWb[sSheet]["content"][iRow] = row
            return True
        except Exception as ex:
            pn(f"Internal problem: Could not set value in multisheet data structure: sheet={sSheet}, header={sHead}, row={iRow}, value={value}, problem:{ex}")
            return False

    def isFinalized(self):
        return self.__bFinalized

    def isIndexEmpty(self):
        for sSheet, dicIndex in self.__dicIndex.items():
            if len(dicIndex) > 0: return False
        return True

    def getNoOfSheets(self):
        return len(self.__odicWb.keys())

    def getSheetNames(self):
        return list(self.__odicWb.keys())

    def getColNames(self, sSheet):
        try:
            return self.__odicWb[sSheet]["header"]
        except:
            return None



